#include <cstring>
#include <cmath>
#include <map>
#include <string>
#include <vector>
#include <cassert>
using namespace std;

#include "graphix.h"
#include <GL/gl.h>
#include <GL/glu.h>
#include <GL/glx.h>



class BitmapManipulator{
	public:
		map<string,pair<GLuint,SDL_Surface*> > filenames;
		pair<GLuint,SDL_Surface*> LoadBitmap(string filename);
} resources;

pair<GLuint,SDL_Surface*> BitmapManipulator::LoadBitmap(string filename){
	map<string,pair<GLuint,SDL_Surface*> >::iterator i=this->filenames.find(filename);
	if(i==this->filenames.end()){
		SDL_Surface *TextureImage[1];
		if((TextureImage[0] = SDL_LoadBMP(filename.c_str()))){
			GLuint bitm;
			glGenTextures(1,&bitm);
			glBindTexture(GL_TEXTURE_2D,bitm);
			glTexImage2D(GL_TEXTURE_2D,0,3,TextureImage[0]->w,TextureImage[0]->h,0,GL_BGR,GL_UNSIGNED_BYTE, TextureImage[0]->pixels);

			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
			glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
	//		if(TextureImage[0])SDL_FreeSurface(TextureImage[0]);
			return pair<GLuint,SDL_Surface*>(bitm,TextureImage[0]);
		}else{
			if(TextureImage[0])SDL_FreeSurface(TextureImage[0]);
			return pair<GLuint,SDL_Surface*>(0,NULL);
		}
	}else{
		return i->second;
	}
	return pair<GLuint,SDL_Surface*>(0,NULL);
}

 #define GL_DRAW_TEXTURE(x1,y1,x2,y2) {glTexCoord2f(0.0f, 0.0f); 	glVertex2f(x1,y1);	glTexCoord2f(1.0f, 0.0f);	glVertex2f(x2,y1 );	glTexCoord2f(1.0f, 1.0f); 	glVertex2f(x2,y2);	glTexCoord2f(0.0f, 1.0f); 	glVertex2f(x1,y2);}

void LoadTexture(Texture &texture,FILE *fp){
//	return;
	char filename[1000],mask[1000];int w,h;
	fscanf(fp,"%s %s %d %d",filename,mask,&w,&h);
	texture.w=w;
	texture.h=h;
	pair<GLuint,SDL_Surface*> res=resources.LoadBitmap(string(filename));
	pair<GLuint,SDL_Surface*> mas=resources.LoadBitmap(string(mask));
	texture.texture[0]=res.first;
	texture.text[0]=res.second;
	texture.mask[0]=mas.first;
	texture.mas[0]=mas.second;
	if(!texture.texture[0])printf("Error while loading texture: %s\n",filename);
	return;	
} 

Graphix::Graphix(int screen_width,int screen_height,int screen_bpp,bool full_screen,bool double_buffer){
	const SDL_VideoInfo *videoInfo;
	this->screen_width=screen_width;
	this->screen_height=screen_height;
	this->screen_bpp=screen_bpp;
	this->full_screen=full_screen;
	this->double_buffer=double_buffer;
	this->background=NULL;
	this->dirt=NULL;
	if(SDL_Init(SDL_INIT_VIDEO|SDL_INIT_NOPARACHUTE)<0){
		fprintf(stderr,"Video initialization failed: %s\n",SDL_GetError());
		created=false;
		return;
	}
	videoInfo=SDL_GetVideoInfo();
	if(!videoInfo){
		fprintf(stderr,"Video query failed: %s\n",SDL_GetError());
		created=false;
		return;
	}
	/* the flags to pass to SDL_SetVideoMode */
	this->flags  = SDL_OPENGL;          /* Enable OpenGL in SDL */
	if(this->double_buffer)flags |= SDL_GL_DOUBLEBUFFER; /* Enable double buffering */
	this->flags |= SDL_HWPALETTE;       /* Store the palette in hardware */
	this->flags |= SDL_RESIZABLE;       /* Enable window resizing */
	if(this->full_screen)this->flags |= SDL_FULLSCREEN;
						    
	/* This checks to see if surfaces can be stored in memory */
//	if(videoInfo->hw_available==SDL_HWSURFACE)printf("hw aviable\n");else printf("oh no\n");
//	this->flags=SDL_HWSURFACE|SDL_ANYFORMAT|SDL_HWPALETTE|SDL_DOUBLEBUF;
//	if(videoInfo->hw_available)this->flags |= SDL_HWSURFACE;
//		else this->flags |= SDL_SWSURFACE;
	/* This checks if hardware blits can be done */
//	if(videoInfo->blit_hw)printf("hw_accel\n");//this->flags |= SDL_HWACCEL;
	SDL_GL_SetAttribute(SDL_GL_DOUBLEBUFFER,1);
	this->surface=SDL_SetVideoMode(this->screen_width,this->screen_height,this->screen_bpp,this->flags);
	videoInfo=SDL_GetVideoInfo();
//	if(videoInfo->hw_available==SDL_HWSURFACE)printf("hw aviable\n");else printf("oh no\n");

	/* Verify there is a surface */
	if (!this->surface){
		fprintf(stderr,"Video mode set failed: %s\n",SDL_GetError());
		created=false;
		return;
	}
	this->initGL();
	this->resize_window(this->screen_width,this->screen_height);
	this->created=true;
	// ideme loadovat resource
	// format: N - pocet bitmap, dales N stringov zo subormi
	// M - pocet hracov, potom M nacitani hracof
/*	FILE *fp=fopen("graphix/files","r");
	int N;
	fscanf(fp,"%d ",&N);
	for(int i=0;i<N;i++){
		char filename[1000];
		fscanf(fp,"%s ",filename);
		pair<GLuint,SDL_Surface*> res=resources.LoadBitmap(string(filename));
		this->mapBitmaps.push_back(res.first);
		this->sMapBitmaps.push_back(res.second);
	}
	int M;
	fscanf(fp,"%d ",&M);
	this->playerGraphix.resize(M);
	for(int i=0;i<M;i++){
		LoadPlayerGraphix(this->playerGraphix[i],fp);
	}
	int O;
	fscanf(fp,"%d ",&O);
	this->objectGraphix.resize(O);
	for(int i=0;i<O;i++){
		LoadObjectGraphix(this->objectGraphix[i],fp);
	}
	fclose(fp);*/
}

Graphix::~Graphix(){
	destroy_font();
	SDL_QuitSubSystem(SDL_INIT_VIDEO);	
}

int	Graphix::load_graphix(string filename){
	printf("%s\n",filename.c_str());
	FILE *fp=fopen(filename.c_str(),"r");
	int M;
	fscanf(fp,"%d ",&M);
	this->textures.resize(M);
	for(int i=0;i<M;i++){
		LoadTexture(this->textures[i],fp);
	}
/*	int O;
	fscanf(fp,"%d ",&O);
	this->objectGraphix.resize(O);
	for(int i=0;i<O;i++){
		LoadObjectGraphix(this->objectGraphix[i],fp);
	}*/
	fclose(fp);
	return 0;
}

int Graphix::free_graphix(){
	for(unsigned int i=0;i<textures.size();i++)textures[i].free();
	textures.resize(0);
	SDL_FreeSurface(background);
	background=NULL;
	SDL_FreeSurface(background2);
	background2=NULL;
	glDeleteTextures(1,&gl_background);		
	glDeleteTextures(1,&gl_background2);
	return 0;	
}


int Graphix::initGL(GLvoid){
	glShadeModel(GL_SMOOTH);
	glClearColor(0.0f,0.0f,0.0f,0.0f);
	glClearDepth(1.0f);
	glEnable(GL_TEXTURE_2D);
	glShadeModel(GL_SMOOTH);
//	glEnable(GL_DEPTH_TEST);
	glEnable(GL_BLEND);
	glDepthFunc(GL_LEQUAL);
	glHint(GL_PERSPECTIVE_CORRECTION_HINT,GL_NICEST);
	this->build_font();
        return 1 ;
}


int Graphix::resize_window(int width,int height){
	GLfloat ratio;
	if(height==0)
	height=1;
	ratio=(GLfloat)width/(GLfloat)height;
	glViewport(0,0,(GLsizei)width,(GLsizei)height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(0,width,height,0,-1,1);
//	gluPerspective(45.0f,ratio,0.1f,100.0f);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	//zglFrustum(0,width,height,0,0,0);
	return 1 ;
}

GLvoid Graphix::build_font(GLvoid){
	Display *dpy;          /* Our current X display */
	XFontStruct *fontInfo; /* Our font info */
        /* Sotrage for 96 characters */
	base=glGenLists(96);
	/* Get our current display long enough to get the fonts */
	dpy=XOpenDisplay(NULL);
	/* Get the font information */
	fontInfo=XLoadQueryFont(dpy,"-adobe-helvetica-medium-r-normal--18-*-*-*-p-*-iso8859-1");
	/* If the above font didn't exist try one that should */
	if(fontInfo==NULL){
		fontInfo=XLoadQueryFont(dpy,"fixed");
		/* If that font doesn't exist, something is wrong */
		if(fontInfo==NULL){
			fprintf(stderr, "no X font available?\n");
		}
	}
	/* generate the list */
	glXUseXFont(fontInfo->fid,32,96,base);
	/* Recover some memory */
	XFreeFont(dpy,fontInfo);
	/* close the display now that we're done with it */
	XCloseDisplay(dpy);
	return;
}

GLvoid Graphix::destroy_font(){
	glDeleteLists(this->base,96);
}

GLvoid Graphix::print(double x,double y,double r,double g,double b,const char *fmt,...){
	glColor3f(r,g,b);
                /* Position The Text On The Screen */
        glRasterPos2f(x,y);
	char text[256]; /* Holds our string */
	va_list ap;     /* Pointer to our list of elements */
	/* If there's no text, do nothing */
	if(fmt==NULL)return;
	/* Parses The String For Variables */
	va_start(ap,fmt);
	/* Converts Symbols To Actual Numbers */
	vsprintf(text,fmt,ap);
	va_end(ap);
	/* Pushes the Display List Bits */
	glPushAttrib(GL_LIST_BIT);
	/* Sets base character to 32 */
	glListBase(base-32);
	/* Draws the text */
	glCallLists(strlen(text),GL_UNSIGNED_BYTE,text);
	/* Pops the Display List Bits */
	glPopAttrib();
	glColor3f(1,1,1);
//	SDL_GL_SwapBuffers();
}

int dim1,dim2;

int get_cool_dimension(int dim){
	//return dim;
	int a=1;
	while(a<dim)a*=2;
	return a;
}

void	Graphix::flush_screen(int x1,int y1,int x2,int y2){
	SDL_GL_SwapBuffers();
	x1=y1=x2=y2;
	//SDL_UpdateRect(this->surface,10,10,this->screen_width,this->screen_height);
	//SDL_UpdateRect(this->surface,x1,y1,x2,y2);
}



void	Graphix::draw_texture(int texture,int x,int y,int w, int h){
		if(w<0)w=this->textures[texture].w;
		if(h<0)h=this->textures[texture].h;
		glEnable(GL_TEXTURE_2D);
		glEnable(GL_BLEND);
		glBlendFunc(GL_DST_COLOR,GL_ZERO);
		glBindTexture(GL_TEXTURE_2D,this->textures[texture].mask[0]);
		glBegin(GL_QUADS);
			GL_DRAW_TEXTURE(x,y,x+w,y+h);
		glEnd();
		glBlendFunc(GL_ONE, GL_ONE);
		glBindTexture(GL_TEXTURE_2D,this->textures[texture].texture[0]);
		glBegin(GL_QUADS);
			GL_DRAW_TEXTURE(x,y,x+w,y+h);
		glEnd();
		glDisable(GL_BLEND);
		glDisable(GL_TEXTURE_2D);
	
}

int progres_bar(int x1,int y1,int x2,int y2,double r,double g,double p){
	glBindTexture(GL_TEXTURE_2D,0);
	glBegin(GL_QUADS);
	glColor3f(1.0f,0.0f,0.0f);
	glVertex2f(x1,y1);
	glColor3f(r,g,0.0f);
	glVertex2f(((double)x1+(double)(x2-x1)*p),y1);
	glVertex2f(((double)x1+(double)(x2-x1)*p),y2);
	glColor3f(1.0f,0.0f,0.0f);
	glVertex2f(x1,y2);
	glEnd();
	return 0;
}


void	Graphix::clear_color(double r,double g,double b){
	glClearColor(r,g,b,0.0);
	glClear(GL_COLOR_BUFFER_BIT);
}
